import torch
import numpy as np
import random
from experiment.main import args
def accuracy(output, labels, test=False):
    preds = output.max(1)[1].type_as(labels)
    correct = preds.eq(labels).double()
    correct = correct.sum()
    if test:
        return correct/len(labels), preds.eq(labels)
    else:
        return correct / len(labels)
def mig_loss_function_multi_W(output_dict, modal_num, p, bs, device,sampling_num):

    '''

    :param output_dict:  bs*C output matrix per value
    :param modal_num: the number of modalities
    :param p: prior
    :param bs: batch size
    :param device: device number

    :return: MIG loss
    '''

    noise = torch.rand(1).to(device) * 0.0001
    normalize_1 = bs
    normalize_2 = sampling_num

    # Multiply prior on the denominator
    output_dict[1] = output_dict[1] / (p ** (modal_num-1))
    m1 = output_dict[1]
    for i in range(2,modal_num+1):
        m1 = m1 * output_dict[i]
    m1 = torch.log(m1.sum(1) + noise).sum()

    m2 = 0
    # Sample margin from mini-batch
    for idx in range(sampling_num):
        idx = torch.randperm(bs)
        o1 = output_dict[1][idx]
        for i in range(2, modal_num + 1):
            idx = torch.randperm(bs)
            o1 = o1 * output_dict[i][idx]
        #o1 = torch.log(o1.sum(1) + noise).sum()
        m2 = m2 + o1.sum()

    return -(m1) / normalize_1 + m2 / normalize_2

def hgr_loss_function(output1, output2, device):
    N = output1.size(0)
    C = output1.size(1)
    I = torch.FloatTensor(np.eye(N), ).to(device)
    output1 = output1 - output1.sum(0)/N
    output2 = output2 - output2.sum(0)/N
    m = (output1 @ output2.transpose(1,0))
    m1 = (m * I).sum() / (N-1)

    cov1 = output1[0].resize(C, 1) @ output1[0].resize(1, C)
    cov2 = output2[0].resize(C, 1) @ output2[0].resize(1, C)
    for i in range(1,N):
        cov1 = cov1 + output1[i].resize(C, 1) @ output1[i].resize(1, C)
        cov2 = cov2 + output2[i].resize(C, 1) @ output2[i].resize(1, C)

    cov1 = cov1 / (N-1)
    cov2 = cov2 / (N-1)

    return -m1 + 0.5 * torch.trace(cov1 @ cov2)


def mig_loss_function(output1, output2, p, bs, device):

    I = torch.FloatTensor(np.eye(bs), ).to(device)
    E = torch.FloatTensor(np.ones((bs, bs))).to(device)
    normalize_1 = bs
    normalize_2 = bs * bs- bs

    new_output = output1 / p
    m =  (new_output @ output2.transpose(1,0))
    noise = torch.rand(1).to(device)*0.0001
    m1 = torch.log(m * I + I * noise + E - I)
    m2 = m * (E-I)


    return -(m1.sum() + bs) / normalize_1 + m2.sum() / normalize_2

def mig_loss_function_U(output1, output2, p, bs, device):

    I = torch.FloatTensor(np.eye(bs), ).to(device)
    E = torch.FloatTensor(np.ones((bs, bs))).to(device)
    normalize_1 = bs
    normalize_2 = bs * bs- bs
    new_output = output1 / p
    m =  1.0 / (new_output @ output2.transpose(1,0)) #1/K
    noise = torch.rand(1).to(device)*0.0001
    m1 = m * I
    m2 = torch.log(m * (E-I) + I + (E-I) * noise)


    return (m1.sum()) / normalize_1 - m2.sum() / normalize_2


def js_loss_function(output1, output2, p, bs):

    I = torch.FloatTensor(np.eye(bs), ).cuda()
    E = torch.FloatTensor(np.ones((bs, bs))).cuda()
    normalize_1 = bs
    normalize_2 = bs * bs - bs

    new_output = output1 / p
    m = (new_output @ output2.transpose(1, 0))
    noise = torch.rand(1).cuda() * 0.0001
    m1 = torch.log( ((2 * m)/(1 + m))* I  + I * noise + E - I)
    m2 = torch.log( (2/(1+m))* (E - I) + (E-I)*noise + I)

    return -(m1.sum()) / normalize_1 - m2.sum() / normalize_2

def js_loss_function_multi(output_dict, modal_num, p, bs, device,sampling_num):

    '''

    :param output_dict:  bs*C output matrix per value
    :param modal_num: the number of modalities
    :param p: prior
    :param bs: batch size
    :param device: device number

    :return: MIG loss
    '''

    noise = torch.rand(1).to(device) * 0.0001
    normalize_1 = bs
    normalize_2 = sampling_num

    # Multiply prior on the denominator
    output_dict[1] = output_dict[1] / (p ** (modal_num-1))
    m1 = output_dict[1]
    for i in range(2,modal_num+1):
        m1 = m1 * output_dict[i]

    m1 = m1.sum(1)
    m1 = 2 * m1 / (1+m1)
    m1 = torch.log(m1 + noise).sum()

    m2 = 0.
    # Sample margin from mini-batch
    for idx in range(sampling_num):
        idx = torch.randperm(bs)
        o1 = output_dict[1][idx]
        for i in range(2, modal_num + 1):
            idx = torch.randperm(bs)
            o1 = o1 * output_dict[i][idx]
        o1 = o1.sum(1)
        o1 = torch.log(2 / (1+o1) + noise).sum()
        m2 = m2 + o1

    return -(m1) / normalize_1 - m2 / normalize_2



def mig_loss_function_multi(output_dict, modal_num, p, bs, device,sampling_num,d):

    '''

    :param output_dict:  bs*C output matrix per value
    :param modal_num: the number of modalities
    :param p: prior
    :param bs: batch size
    :param device: device number

    :return: MIG loss
    '''

    noise = torch.rand(1).to(device) * 0.0001
    normalize_1 = bs
    normalize_2 = len(d[0])

    # Multiply prior on the denominator
    output_dict[1] = output_dict[1] / (p ** (modal_num-1))

    # Sample joint from mini-batch
    m1 = output_dict[1]
    for i in range(2,modal_num+1):
        m1 = m1 * output_dict[i]
    m1 = torch.log(m1.sum(1) + noise)
    m1 = m1.sum()

    # Sample margin from mini-batch
    o1 = output_dict[1][d[0]]
    for i in range(2, modal_num + 1):

        o1 = o1 * output_dict[i][d[i-1]]

    o1 = o1.sum(1).sum()
    m2 = o1

    return m1 / normalize_1 - m2 / normalize_2


def enu_modalities(m_num):
    p = []
    for i in range(1,m_num+1):
        for j in range(i+1,m_num+1):
            p += [[i,j]]

    return p
def get_index_3(bs):
    p = []
    for i in range(0,bs):
        for j in range(0,bs):
            if i == j:
                continue
            for k in range(0,bs):
                if i==k or j==k:
                    continue
                p += [[i,j,k]]

    index_1 = []
    index_2 = []
    index_3 = []

    #print(p[0],len(p))
    for i in range(len(p)):
        index_1 += [p[i][0]]
        index_2 += [p[i][1]]
        index_3 += [p[i][2]]



    return [index_1,index_2,index_3]
def get_index_6(bs):
    p = []
    for i in range(0,bs):
        for j in range(0,bs):
            if i == j:
                continue
            for k in range(0,bs):
                if i==k or j==k:
                    continue
                for a in range(0,bs):
                    if a == i or a == j or a==k:
                        continue
                    for b in range(0,bs):
                        if b == i or b == j or b == k or b == a:
                            continue

                        p += [[i,j,k,a,b]]
    sample = 50000
    index_1 = []
    index_2 = []
    index_3 = []
    index_4 = []
    index_5 = []
    #print(p[0],len(p))
    for i in range(len(p)):
        index_1 += [p[i][0]]
        index_2 += [p[i][1]]
        index_3 += [p[i][2]]
        index_4 += [p[i][3]]
        index_5 += [p[i][4]]

    index_1 = index_1[:sample]
    index_2 = index_2[:sample]
    index_3 = index_3[:sample]
    index_4 = index_4[:sample]
    index_5 = index_5[:sample]


    return [index_1,index_2,index_3,index_4,index_5]

